# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generate config files for Noto-COLRv1

from nanoemoji.util import rel
from pathlib import Path


_NOTO_FAMILY_NAME = "Noto Color Emoji"
_NOTO_SVG_DIR = Path("svg")
_REGIONAL_INDICATORS = {
    Path(_NOTO_SVG_DIR / f"emoji_u{i:x}.svg") for i in range(0x1F1E6, 0x1F1FF + 1)
}
_NOTO_WAVED_FLAG_SVG_DIR = Path("third_party/region-flags/waved-svg")
_NOTO_SUBDIVISION_FLAGS = (
    _NOTO_WAVED_FLAG_SVG_DIR / "emoji_u1f3f4_e0067_e0062_e0065_e006e_e0067_e007f.svg",
    _NOTO_WAVED_FLAG_SVG_DIR / "emoji_u1f3f4_e0067_e0062_e0073_e0063_e0074_e007f.svg",
    _NOTO_WAVED_FLAG_SVG_DIR / "emoji_u1f3f4_e0067_e0062_e0077_e006c_e0073_e007f.svg",
)
_CONFIG_DIR = Path("colrv1")


def _write_config(config_name, output_file, svg_files):
    svg_files = [rel(_CONFIG_DIR, Path(f)) for f in svg_files]
    config_file = f"{config_name}.toml"
    for svg_file in svg_files:
        assert _CONFIG_DIR.joinpath(
            svg_file
        ).is_file(), f"{svg_file} not found relative to {_CONFIG_DIR}"
    svg_list = ",\n    ".join(f'"{f}"' for f in sorted(svg_files)).rstrip()
    with open(_CONFIG_DIR / config_file, "w") as f:
        f.write(
            f"""
family = "{_NOTO_FAMILY_NAME}"
output_file = "{output_file}"
color_format = "glyf_colr_1"
clipbox_quantization = 32

[axis.wght]
name = "Weight"
default = 400

[master.regular]
style_name = "Regular"
srcs = [
    {svg_list},
]

[master.regular.position]
wght = 400
"""
        )


def _write_all_noto_configs():
    # includes all of noto-emoji svgs plus all the waved region flags
    regular = tuple(_NOTO_SVG_DIR.glob("*.svg"))
    flags = tuple(_NOTO_WAVED_FLAG_SVG_DIR.glob("*.svg"))

    dups = {p.name for p in regular} & {p.name for p in flags}
    if dups:
        raise ValueError(f"Flags *and* regular have {dups}")

    svgs = regular + flags
    _write_config("all", "NotoColorEmoji.ttf", svgs)


def _write_noto_noflag_configs():
    # Does not contain regional indicators and the region flags that are
    # composed with them. It still includes the England, Scotland and Wales
    # 'subdivision' flags, as those are not composed with Unicode regional
    # indicators, but use sequences of Unicode Tag letters prefixed with
    # the Black Flag and ending with a Cancel Tag.
    svgs = (
        tuple(p for p in _NOTO_SVG_DIR.glob("*.svg") if p not in _REGIONAL_INDICATORS)
        + _NOTO_SUBDIVISION_FLAGS
    )
    _write_config("noflags", "NotoColorEmoji-noflags.ttf", svgs)


def main():
    _write_all_noto_configs()
    _write_noto_noflag_configs()


if __name__ == "__main__":
    main()



from fontTools import ttLib
from pathlib import Path
import pytest
import re


NAME_ID_FAMILY = 1
NAME_ID_UNIQUE_ID = 3
NAME_ID_FULLNAME = 4
NAME_ID_POSTSCRIPT_NAME = 6


def test_consistent_version():
    fonts_dir = Path("fonts")
    assert fonts_dir.is_dir()

    name5_re = re.compile(r"^Version (\d+.\d+);GOOG;noto-emoji:\d+:[a-z0-9]+$")

    debug_versions = []
    versions = set()
    for font_file in fonts_dir.rglob("*.ttf"):
        font = ttLib.TTFont(font_file)
        head_ver = f"{font['head'].fontRevision:.03f}"
        versions.add(head_ver)
        debug_versions.append(f"{font_file.name} head {head_ver}")
        for name in font["name"].names:
            # name 5 is version
            if name.nameID != 5:
                continue
            if not name.isUnicode():
                continue
            match = name5_re.match(name.toUnicode())
            assert match is not None, f"{name.toUnicode()} is malformed"
            versions.add(match.group(1))
            debug_versions.append(f"{font_file.name} name {match.group(1)}")
    debug_versions = "\n".join(debug_versions)
    assert (
        len(versions) == 1
    ), f"Should have a consistent version, found\n{debug_versions}"


def test_consistent_fstype():
    fonts_dir = Path("fonts")
    assert fonts_dir.is_dir()

    name5_re = re.compile(r"^Version (\d+.\d+);GOOG;noto-emoji:\d+:[a-z0-9]+$")

    debug_fstypes = []
    fstypes = set()
    for font_file in fonts_dir.rglob("*.ttf"):
        font = ttLib.TTFont(font_file)
        fstype = font["OS/2"].fsType
        fstypes.add(fstype)
        debug_fstypes.append(f"{font_file.name} fsType {fstype}")
    debug_fstypes = "\n".join(debug_fstypes)
    assert fstypes == {0}, f"All fsType's should be 0, found\n{debug_fstypes}"


def test_has_emojicompat():
    fonts_dir = Path("fonts")
    assert fonts_dir.is_dir()

    ec_fonts = set(fonts_dir.rglob("*-emojicompat.ttf"))
    assert {f.name for f in ec_fonts} == {
        "Noto-COLRv1-emojicompat.ttf",
        "NotoColorEmoji-emojicompat.ttf",
    }

    for font_file in ec_fonts:
        font = ttLib.TTFont(font_file)
        assert "meta" in font, f"{font_file.name} should have a meta table"
        assert (
            "Emji" in font["meta"].data
        ), f"{font_file.name} should have emojicompat data"


def name(font, name_id):
    values = set()
    for name in font["name"].names:
        if name.nameID == name_id:
            values.add(name.toUnicode())
    assert len(values) == 1, f"{name_id} has multiple definitions: {values}"
    return next(iter(values))


def test_flagsonly_name():
    fonts_dir = Path("fonts")
    assert fonts_dir.is_dir()
    font_file = fonts_dir / "NotoColorEmoji-flagsonly.ttf"
    font = ttLib.TTFont(font_file)
    assert [
        "Noto Color Emoji Flags",
        "Noto Color Emoji Flags",
        "Noto Color Emoji Flags",
        "NotoColorEmojiFlags",
    ] == [
        name(font, NAME_ID_FAMILY),
        name(font, NAME_ID_FULLNAME),
        name(font, NAME_ID_UNIQUE_ID),
        name(font, NAME_ID_POSTSCRIPT_NAME),
    ]



"""
Post-nanoemoji processing of the Noto COLRv1 Emoji files.

Adds additional sequences to properly support Safari, corrects 'name', etc.

For now substantially based on copying from a correct bitmap build.
"""
from absl import app
import functools
from fontTools.feaLib.builder import addOpenTypeFeaturesFromString
from fontTools import ttLib
from fontTools.ttLib.tables import _g_l_y_f as glyf
from fontTools.ttLib.tables import otTables as ot
import map_pua_emoji
from nototools import add_vs_cmap
from nototools import font_data
from nototools import unicode_data
from pathlib import Path
import re

from colrv1_add_soft_light_to_flags import add_soft_light_to_flags


_CBDT_FILE = Path("fonts/NotoColorEmoji.ttf")
_COLR_FILES = {
    Path("fonts/Noto-COLRv1-noflags.ttf"),
    Path("fonts/Noto-COLRv1.ttf"),
}


def _is_colrv1(font):
    return "COLR" in font and font["COLR"].version == 1


def _is_cbdt(font):
    return "CBDT" in font


def _set_name(name_table, nameID):
    name_table.getName(value, nameID, 3, 1, 0x409)


def _set_name(name_table, nameID, value):
    name_table.setName(value, nameID, 3, 1, 0x409)


def _copy_names(colr_font, cbdt_font):
    colr_font["name"] = cbdt_font["name"]
    name_table = colr_font["name"]
    assert all(
        (n.platformID, n.platEncID, n.langID) == (3, 1, 0x409) for n in name_table.names
    ), "Should only have names Android uses"

    # Amendments
    _set_name(name_table, 10, "Color emoji font using COLRv1.")
    _set_name(name_table, 11, "https://github.com/googlefonts/noto-emoji")
    _set_name(name_table, 12, "https://github.com/googlefonts/noto-emoji")


# CBDT build step: @$(VS_ADDER) -vs 2640 2642 2695 --dstdir '.' -o "$@-with-pua-varsel" "$@-with-pua"
def _add_vs_cmap(colr_font):
    emoji_variants = unicode_data.get_unicode_emoji_variants() | {
        0x2640,
        0x2642,
        0x2695,
    }
    add_vs_cmap.modify_font("COLRv1 Emoji", colr_font, "emoji", emoji_variants)


def _is_variation_selector_cmap_table(table):
    assert table.format in {4, 12, 14}
    return table.format == 14


def _lookup_in_cmap(colr_font, codepoint):
    result = set()
    for table in colr_font["cmap"].tables:
        if _is_variation_selector_cmap_table(table):
            continue
        assert codepoint in table.cmap
        result.add(table.cmap[codepoint])
    assert len(result) == 1, f"Ambiguous mapping for {codepoint}: {result}"
    return next(iter(result))


def _add_cmap_entries(colr_font, codepoint, glyph_name):
    for table in colr_font["cmap"].tables:
        if _is_variation_selector_cmap_table(table):
            continue
        if not _is_bmp(codepoint) and table.format == 4:
            continue
        table.cmap[codepoint] = glyph_name
        #print(f"Map 0x{codepoint:04x} to {glyph_name}, format {table.format}")


FLAG_TAGS = set(range(0xE0030, 0xE0039 + 1)) | set(range(0xE0061, 0xE007A + 1))
CANCEL_TAG = 0xE007F


def _map_missing_flag_tag_chars_to_empty_glyphs(colr_font):
    # Add all tag characters used in flags + cancel tag
    tag_cps = FLAG_TAGS | {CANCEL_TAG}

    # Anything already cmap'd is fine
    tag_cps -= set(_Cmap(colr_font).keys())

    # CBDT maps these to blank glyphs
    glyf_table = colr_font["glyf"]
    hmtx_table = colr_font["hmtx"]
    glyph_order_size = len(glyf_table.glyphOrder)
    for cp in tag_cps:
        #print(f"Map 0x{cp:04x} to a blank glyf")
        glyph_name = f"u{cp:04X}"
        assert glyph_name not in glyf_table, f"{glyph_name} already in glyf"
        assert glyph_name not in hmtx_table.metrics, f"{glyph_name} already in hmtx"
        glyf_table[glyph_name] = glyf.Glyph()
        hmtx_table[glyph_name] = (0, 0)

        _add_cmap_entries(colr_font, cp, glyph_name)


def _is_bmp(cp):
    return cp in range(0x0000, 0xFFFF + 1)


def _ligaset_for_glyph(lookup_list, glyph_name):
    for lookup in lookup_list.Lookup:
        if lookup.LookupType != 4:
            continue
        for liga_set in lookup.SubTable:
            if glyph_name in liga_set.ligatures:
                return liga_set.ligatures[glyph_name]
    return None


def _Cmap(ttfont):
    def _Reducer(acc, u):
        acc.update(u)
        return acc

    unicode_cmaps = (t.cmap for t in ttfont["cmap"].tables if t.isUnicode())
    return functools.reduce(_Reducer, unicode_cmaps, {})


def _add_vertical_layout_tables(cbdt_font, colr_font):
    upem_scale = colr_font["head"].unitsPerEm / cbdt_font["head"].unitsPerEm

    vhea = colr_font["vhea"] = ttLib.newTable("vhea")
    vhea.tableVersion = 0x00010000
    vhea.ascent = round(cbdt_font["vhea"].ascent * upem_scale)
    vhea.descent = round(cbdt_font["vhea"].descent * upem_scale)
    vhea.lineGap = 0
    # most of the stuff below is recalculated by the compiler, but still needs to be
    # initialized... ¯\_(ツ)_/¯
    vhea.advanceHeightMax = 0
    vhea.minTopSideBearing = 0
    vhea.minBottomSideBearing = 0
    vhea.yMaxExtent = 0
    vhea.caretSlopeRise = 0
    vhea.caretSlopeRun = 0
    vhea.caretOffset = 0
    vhea.reserved0 = 0
    vhea.reserved1 = 0
    vhea.reserved2 = 0
    vhea.reserved3 = 0
    vhea.reserved4 = 0
    vhea.metricDataFormat = 0
    vhea.numberOfVMetrics = 0

    # emoji font is monospaced -- except for an odd uni0000 (NULL) glyph which happens
    # to have height=0; but colrv1 font doesn't have that anyway, so I just skip it
    cbdt_heights = set(h for h, _ in cbdt_font["vmtx"].metrics.values() if h != 0)
    assert len(cbdt_heights) == 1, "NotoColorEmoji CBDT font should be monospaced!"
    height = round(cbdt_heights.pop() * upem_scale)
    vmtx = colr_font["vmtx"] = ttLib.newTable("vmtx")
    vmtx.metrics = {}
    for gn in colr_font.getGlyphOrder():
        vmtx.metrics[gn] = height, 0


UNKNOWN_FLAG_PUA = 0xFE82B
BLACK_FLAG = 0x1F3F4
REGIONAL_INDICATORS = set(range(0x1F1E6, 0x1F1FF + 1))


def _add_fallback_subs_for_unknown_flags(colr_font):
    """Add GSUB lookups to replace unsupported flag sequences with the 'unknown flag'.

    In order to locate the unknown flag, the glyph must be mapped to 0xFE82B PUA code;
    the latter is removed from the cmap table after the GSUB has been updated.
    """
    cmap = _Cmap(colr_font)
    unknown_flag = cmap[UNKNOWN_FLAG_PUA]
    black_flag = cmap[BLACK_FLAG]
    cancel_tag = cmap[CANCEL_TAG]
    flag_tags = sorted(cmap[cp] for cp in FLAG_TAGS)
    # in the *-noflags.ttf font there are no region flags thus this list is empty
    regional_indicators = sorted(cmap[cp] for cp in REGIONAL_INDICATORS if cp in cmap)

    classes = f'@FLAG_TAGS = [{" ".join(flag_tags)}];\n'
    if regional_indicators:
        classes += f"""
            @REGIONAL_INDICATORS = [{" ".join(regional_indicators)}];
            @UNKNOWN_FLAG = [{" ".join([unknown_flag] * len(regional_indicators))}];
        """
    lookups = (
        # the first lookup is a dummy that stands for the emoji sequences ligatures
        # from the destination font; we only use it to ensure the lookup indices match.
        # We can't leave it empty otherwise feaLib optimizes it away.
        f"""
        lookup placeholder {{
            sub {unknown_flag} {unknown_flag} by {unknown_flag};
        }} placeholder;
        """
        + "\n".join(
            ["lookup delete_glyph {"]
            + [f"    sub {g} by NULL;" for g in sorted(regional_indicators + flag_tags)]
            + ["} delete_glyph;"]
        )
        + (
            """
            lookup replace_with_unknown_flag {
                sub @REGIONAL_INDICATORS by @UNKNOWN_FLAG;
            } replace_with_unknown_flag;
            """
            if regional_indicators
            else "\n"
        )
    )
    features = (
        "languagesystem DFLT dflt;\n"
        + classes
        + lookups
        + "feature ccmp {"
        + f"""
            lookup placeholder;
            sub {black_flag} @FLAG_TAGS' lookup delete_glyph;
            sub {black_flag} {cancel_tag} by {unknown_flag};
        """
        + (
            """
            sub @REGIONAL_INDICATORS' lookup replace_with_unknown_flag
                @REGIONAL_INDICATORS' lookup delete_glyph;
            """
            if regional_indicators
            else ""
        )
        + "} ccmp;"
    )
    # feaLib always builds a new GSUB table (can't update one in place) so we have to
    # use an empty TTFont and then update our GSUB with the newly built lookups
    temp_font = ttLib.TTFont()
    temp_font.setGlyphOrder(colr_font.getGlyphOrder())

    addOpenTypeFeaturesFromString(temp_font, features)

    temp_gsub = temp_font["GSUB"].table
    # sanity check
    assert len(temp_gsub.FeatureList.FeatureRecord) == 1
    assert temp_gsub.FeatureList.FeatureRecord[0].FeatureTag == "ccmp"
    temp_ccmp = temp_gsub.FeatureList.FeatureRecord[0].Feature

    colr_gsub = colr_font["GSUB"].table
    ccmps = [
        r.Feature for r in colr_gsub.FeatureList.FeatureRecord if r.FeatureTag == "ccmp"
    ]
    assert len(ccmps) == 1, f"expected only 1 'ccmp' feature record, found {len(ccmps)}"
    colr_ccmp = ccmps[0]

    colr_lookups = colr_gsub.LookupList.Lookup
    assert (
        len(colr_lookups) == 1
    ), f"expected only 1 lookup in COLRv1's GSUB.LookupList, found {len(colr_lookups)}"
    assert (
        colr_lookups[0].LookupType == 4
    ), f"expected Lookup[0] of type 4 in COLRv1, found {colr_lookups[0].LookupType}"

    colr_lookups.extend(temp_gsub.LookupList.Lookup[1:])
    colr_gsub.LookupList.LookupCount = len(colr_lookups)
    colr_ccmp.LookupListIndex = temp_ccmp.LookupListIndex
    colr_ccmp.LookupCount = len(colr_ccmp.LookupListIndex)

    # get rid of the Unknown Flag private codepoint as no longer needed
    font_data.delete_from_cmap(colr_font, [UNKNOWN_FLAG_PUA])


def _set_no_font_embedding_restrictions(colr_font):
    # The CBDT/CBLC NotoColorEmoji has OS/2.fsType = 0 (i.e. no embedding restrictions)
    # so the COLRv1 variant must also have no such restrictions.
    # https://github.com/notofonts/noto-fonts/issues/2408
    # https://github.com/google/fonts/issues/5729
    colr_font["OS/2"].fsType = 0


def _set_head_version_to_name_version(colr_font):
    # head.fontRevision and the version on name 5 should match
    name_version = colr_font['name'].getName(5, 3, 1, 0x409)
    assert name_version is not None, "No version found in 'name'"
    name_version = name_version.toUnicode()
    match = re.match(r'^Version (\d+[.]\d+);GOOG;', name_version)
    assert match is not None, f"Unable to parse version from '{name_version}'"
    colr_font["head"].fontRevision = float(match.group(1))


def _font(path, check_fn, check_fail_str):
    assert path.is_file(), path
    font = ttLib.TTFont(path)
    if not check_fn(font):
        raise ValueError(path + check_fail_str)
    return font


def main(_):
    cbdt_font = _font(_CBDT_FILE, _is_cbdt, " must be a CBDT font")

    for colr_file in _COLR_FILES:
        colr_font = _font(colr_file, _is_colrv1, " must be a COLRv1 font")

        print(f"Updating {colr_file} from {_CBDT_FILE}")

        _copy_names(colr_font, cbdt_font)

        # CBDT build step: @$(PYTHON) $(PUA_ADDER) "$@" "$@-with-pua"
        map_pua_emoji.add_pua_cmap_to_font(colr_font)

        _add_vs_cmap(colr_font)

        _map_missing_flag_tag_chars_to_empty_glyphs(colr_font)

        add_soft_light_to_flags(colr_font)

        _add_vertical_layout_tables(cbdt_font, colr_font)

        _add_fallback_subs_for_unknown_flags(colr_font)

        _set_no_font_embedding_restrictions(colr_font)

        _set_head_version_to_name_version(colr_font)

        print("Writing", colr_file)
        colr_font.save(colr_file)


if __name__ == "__main__":
    app.run(main)



"""Utility to add soft-light effect to NotoColorEmoji-COLRv1 region flags."""
import sys
import subprocess
from fontTools import ttLib
from fontTools.ttLib.tables import otTables as ot
from fontTools.ttLib.tables.C_P_A_L_ import Color
from fontTools.colorLib.builder import LayerListBuilder
from add_aliases import read_default_emoji_aliases
from flag_glyph_name import flag_code_to_glyph_name
from map_pua_emoji import get_glyph_name_from_gsub


REGIONAL_INDICATOR_RANGE = range(0x1F1E6, 0x1F1FF + 1)
BLACK_FLAG = 0x1F3F4
CANCEL_TAG = 0xE007F
TAG_RANGE = range(0xE0000, CANCEL_TAG + 1)


def is_flag(sequence):
    # regular region flags are comprised of regional indicators
    if all(cp in REGIONAL_INDICATOR_RANGE for cp in sequence):
        return True

    # subdivision flags start with  black flag, contain some tag characters and end with
    # a cancel tag
    if (
        len(sequence) > 2
        and sequence[0] == BLACK_FLAG
        and sequence[-1] == CANCEL_TAG
        and all(cp in TAG_RANGE for cp in sequence[1:-1])
    ):
        return True

    return False


def read_makefile_variable(var_name):
    # see `print-%` command in Makefile
    value = subprocess.run(
        ["make", f"print-{var_name}"], capture_output=True, check=True
    ).stdout.decode("utf-8")
    return value[len(var_name) + len(" = ") :].strip()


def flag_code_to_sequence(flag_code):
    # I use the existing code to first convert from flag code to glyph name,
    # and then convert names back to integer codepoints since it already
    # handles both the region indicators and subdivision tags.
    name = flag_code_to_glyph_name(flag_code)
    assert name.startswith("u")
    return tuple(int(v, 16) for v in name[1:].split("_"))


def all_flag_sequences():
    """Return the set of all noto-emoji's region and subdivision flag sequences.
    These include those in SELECTED_FLAGS Makefile variable plus those listed
    in the 'emoji_aliases.txt' file.
    """
    result = {
        flag_code_to_sequence(flag_code)
        for flag_code in read_makefile_variable("SELECTED_FLAGS").split()
    }
    result.update(seq for seq in read_default_emoji_aliases() if is_flag(seq))
    return result


_builder = LayerListBuilder()


def _build_paint(source):
    return _builder.buildPaint(source)


def _paint_composite(source, mode, backdrop):
    return _build_paint(
        {
            "Format": ot.PaintFormat.PaintComposite,
            "SourcePaint": source,
            "CompositeMode": mode,
            "BackdropPaint": backdrop,
        }
    )


def _palette_index(cpal, color):
    assert len(cpal.palettes) == 1
    palette = cpal.palettes[0]
    try:
        i = palette.index(color)
    except ValueError:
        i = len(palette)
        palette.append(color)
        cpal.numPaletteEntries += 1
        assert len(palette) == cpal.numPaletteEntries
    return i


WHITE = Color.fromHex("#FFFFFFFF")
GRAY = Color.fromHex("#808080FF")
BLACK = Color.fromHex("#000000FF")


def _soft_light_gradient(cpal):
    return _build_paint(
        {
            "Format": ot.PaintFormat.PaintLinearGradient,
            "ColorLine": {
                "Extend": "pad",
                "ColorStop": [
                    {
                        "StopOffset": 0.0,
                        "PaletteIndex": _palette_index(cpal, WHITE),
                        "Alpha": 0.5,
                    },
                    {
                        "StopOffset": 0.5,
                        "PaletteIndex": _palette_index(cpal, GRAY),
                        "Alpha": 0.5,
                    },
                    {
                        "StopOffset": 1.0,
                        "PaletteIndex": _palette_index(cpal, BLACK),
                        "Alpha": 0.5,
                    },
                ],
            },
            "x0": 47,
            "y0": 790,
            "x1": 890,
            "y1": -342,
            "x2": -1085,
            "y2": -53,
        },
    )


def flag_ligature_glyphs(font):
    """Yield ligature glyph names for all the region/subdivision flags in the font."""
    for flag_sequence in all_flag_sequences():
        flag_name = get_glyph_name_from_gsub(flag_sequence, font)
        if flag_name is not None:
            yield flag_name


def add_soft_light_to_flags(font, flag_glyph_names=None):
    """Add soft-light effect to region and subdivision flags in CORLv1 font."""
    if flag_glyph_names is None:
        flag_glyph_names = flag_ligature_glyphs(font)

    colr_glyphs = {
        rec.BaseGlyph: rec
        for rec in font["COLR"].table.BaseGlyphList.BaseGlyphPaintRecord
    }
    cpal = font["CPAL"]

    for flag_name in flag_glyph_names:
        flag = colr_glyphs[flag_name]
        flag.Paint = _paint_composite(
            source=_paint_composite(
                source=_soft_light_gradient(cpal),
                mode=ot.CompositeMode.SRC_IN,
                backdrop=flag.Paint,
            ),
            mode=ot.CompositeMode.SOFT_LIGHT,
            backdrop=flag.Paint,
        )


def main():
    try:
        input_file, output_file = sys.argv[1:]
    except ValueError:
        print("usage: colrv1_add_soft_light_to_flags.py INPUT_FONT OUTPUT_FONT")
        return 2

    font = ttLib.TTFont(input_file)

    if "COLR" not in font or font["COLR"].version != 1:
        print("error: missing required COLRv1 table")
        return 1

    add_soft_light_to_flags(font)

    font.save(output_file)


if __name__ == "__main__":
    sys.exit(main())



"""Set COLRv1 fontRevision from CBDT.

Used for bugfix, should fix to set properly on build instead.
"""

from fontTools import ttLib
from pathlib import Path
import sys


NAME_ID_VERSION = 5


def name(font, name_id):
    return ",".join(n.toUnicode() for n in font["name"].names if n.isUnicode() and n.nameID == name_id)


def main():
    colr_font_files = sorted(p for p in (Path(__file__).parent / "fonts").iterdir() if p.name.startswith("Noto-COLRv1"))

    for colr_font_file in colr_font_files:
        cbdt_font_file = colr_font_file.with_stem(colr_font_file.stem.replace("Noto-COLRv1", "NotoColorEmoji"))

        colr_font = ttLib.TTFont(colr_font_file)
        cbdt_font = ttLib.TTFont(cbdt_font_file)

        assert "CBDT" in cbdt_font
        assert "COLR" in colr_font

        colr_font["head"].fontRevision = cbdt_font["head"].fontRevision

        colr_font.save(colr_font_file)


if __name__ == '__main__':
  main()



"""Removes regional indicators from a font."""

from fontTools import subset
from fontTools import ttLib
import functools
from pathlib import Path
import sys
from typing import Set


def codepoints(font: ttLib.TTFont) -> Set[int]:
    unicode_cmaps = (t.cmap.keys() for t in font['cmap'].tables if t.isUnicode())
    return functools.reduce(lambda acc, u: acc | u, unicode_cmaps, set())


def is_regional_indicator(cp: int) -> bool:
    return 0x1F1E6 <= cp <= 0x1F1FF


def main(argv):
    for font_file in sorted(argv[1:]):
        font_file = Path(font_file)
        assert font_file.is_file(), font_file
        noflags_file = font_file.with_stem(font_file.stem + "-noflags")

        if noflags_file.is_file():
            print(font_file, "already has", noflags_file, "; nop")
            continue

        font = ttLib.TTFont(font_file)

        cps = codepoints(font)
        cps_without_flags = {cp for cp in cps if not is_regional_indicator(cp)}

        if cps == cps_without_flags:
            print(font_file, "has no regional indicators")
            continue

        subsetter = subset.Subsetter()
        subsetter.populate(unicodes=cps_without_flags)
        subsetter.subset(font)

        font.save(noflags_file)
        print(font_file, "=>" , noflags_file)


if __name__ == '__main__':
  main(sys.argv)



"""Prints info about emoji fonts."""

from fontTools import ttLib
from pathlib import Path
import sys


NAME_ID_VERSION = 5


def name(font, name_id):
    return ",".join(n.toUnicode() for n in font["name"].names if n.isUnicode() and n.nameID == name_id)


def main():
    font_files = sorted(p for p in (Path(__file__).parent / "fonts").iterdir() if p.suffix == ".ttf")
    max_name_len = max(len(p.name) for p in font_files)

    for font_file in font_files:
        font = ttLib.TTFont(font_file)

        font_type = []
        if "CBDT" in font:
            font_type.append("CBDT")
        if "COLR" in font:
            font_type.append("COLR")
        if "meta" in font and "Emji" in font["meta"].data:
            font_type.append("EmojiCompat")
        font_type.append(f"fontRevision:{font['head'].fontRevision:.3f}")
        font_type.append(name(font, NAME_ID_VERSION))
        font_type = ", ".join(font_type)

        print(f"{font_file.name:{max_name_len + 1}} {font_type}")


if __name__ == '__main__':
  main()
